//$Id: SprChromosome.hh,v 1.1 2008/02/22 23:07:50 narsky Exp $
// File and Version Information:
//
//
// Description:

//
// Environment:
//      Software developed at Caltech's Center for Advanced Computing Research
//
// Author List:
//      Julian Bunn                     Original author
//
// Copyright Information:
//      Copyright (C) 2008              California Institute of Technology
//
//------------------------------------------------------------------------
 
#ifndef _SprGEPChromosome_HH
#define _SprGEPChromosome_HH

#include "StatPatternRecognition/SprDefs.hh"
#include "StatPatternRecognition/SprClass.hh"


#include <string>
#include <iostream>
#include <vector>

class SprGene {
	public:
		SprGene() {}
		SprGene(const SprGene& other)
			:
			head_(other.head_),
			tail_(other.tail_)
			{}
		SprGene& operator=(const SprGene& other) {
			head_ = other.head_;
			tail_ = other.tail_;
			return *this;
		}
		void setHead(std::vector<int> head) {head_ = head;}
		void setTail(std::vector<int> tail) {tail_ = tail;}
		std::vector<int> getHead() {return head_;}
		std::vector<int> getTail() {return tail_;}
		std::vector<std::vector<int> > generateTree() const;

		void print() const;
	private:
		std::vector<int> head_;
		std::vector<int> tail_;
};

class SprChromosome {
	public:
		std::vector<SprGene> gene_;
		std::vector<double> constants_;
		char link_function_;
		unsigned num_genes_;
		double fitness_;
		unsigned length_;
		
		SprChromosome();
		SprChromosome(unsigned num_genes,
			char link_function);

		double Evaluate(const std::vector<double> &data, int verbose) const;
		void generateCode();

		bool mutate(std::vector<int> funcs, unsigned dim, unsigned constants_per_chromosome);
		bool RIS();
		bool IS();
		bool OnePoint(SprChromosome &other);
		bool TwoPoint(SprChromosome &other);
		bool WholeGene(SprChromosome &other);
		bool ConstantMutation(double range);
		bool ConstantSwap(SprChromosome &other);
		void setLength(unsigned num_genes, unsigned num_constants, unsigned length) {
			num_genes_ = num_genes;
			gene_.resize(num_genes);
			constants_.resize(num_constants);
			length_ = length;
		}

		void print() const;

	private:

};

#endif
